import abc
import warnings

from kani.ai_function import AIFunction
from kani.models import ChatMessage


class BaseCompletion(abc.ABC):
    """Base class for all LM engine completions."""

    @property
    @abc.abstractmethod
    def message(self) -> ChatMessage:
        """The message returned by the LM."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def prompt_tokens(self) -> int | None:
        """How many tokens are in the prompt. Can be None for kani to estimate using tokenizer."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def completion_tokens(self) -> int | None:
        """How many tokens are in the completion. Can be None for kani to estimate using tokenizer."""
        raise NotImplementedError


class Completion(BaseCompletion):
    def __init__(self, message: ChatMessage, prompt_tokens: int | None = None, completion_tokens: int | None = None):
        self._message = message
        self._prompt_tokens = prompt_tokens
        self._completion_tokens = completion_tokens

    @property
    def message(self):
        return self._message

    @property
    def prompt_tokens(self):
        return self._prompt_tokens

    @property
    def completion_tokens(self):
        return self._completion_tokens


class BaseEngine(abc.ABC):
    """Base class for all LM engines.

    To add support for a new LM, make a subclass of this and implement the abstract methods below.
    """

    # ==== required interface ====
    max_context_size: int
    """The maximum context size supported by this engine's LM."""

    @abc.abstractmethod
    def message_len(self, message: ChatMessage) -> int:
        """Return the length, in tokens, of the given chat message."""
        raise NotImplementedError

    @abc.abstractmethod
    async def predict(
        self, messages: list[ChatMessage], functions: list[AIFunction] | None = None, **hyperparams
    ) -> BaseCompletion:
        """
        Given the current context of messages and available functions, get the next predicted chat message from the LM.

        :param messages: The messages in the current chat context. ``sum(message_len(m) for m in messages)`` is
            guaranteed to be less than max_context_size.
        :param functions: The functions the LM is allowed to call.
        :param hyperparams: Any additional parameters to pass to the engine.
        """
        raise NotImplementedError

    # ==== optional interface ====
    token_reserve: int = 0
    """Optional: The number of tokens to reserve for internal engine mechanisms (e.g. if an engine has to set up the
    model's reply with a delimiting token).
    
    Default: 0
    """

    def function_token_reserve(self, functions: list[AIFunction]) -> int:
        """Optional: How many tokens are required to build a prompt to expose the given functions to the model.

        Default: If this is not implemented and the user passes in functions, log a warning that the engine does not
        support function calling.
        """
        if functions:
            warnings.warn(
                f"The {type(self).__name__} engine is conversational only and does not support function calling.\n"
                "Developers: If this warning is incorrect, please implement `function_token_reserve()`."
            )
        return 0

    async def close(self):
        """Optional: Clean up any resources the engine might need."""
        pass



from kani import ChatMessage


def dict_at_least(outer: dict, inner: dict) -> bool:
    """Does the outer dict contain at least all kvs from the inner dict (checking dicts recursively)?"""
    for k, v in inner.items():
        if k not in outer:
            return False
        if not isinstance(outer[k], type(v)):
            return False
        # value: if it's a dict, recursive call; otherwise check equality
        if isinstance(v, dict):
            if not dict_at_least(outer[k], v):
                return False
        elif outer[k] != v:
            return False
    return True


def flatten_chatmessages(msgs: list[ChatMessage], between: str = "") -> str:
    """Return a string that is the concatenation of all contents of the chat messages."""
    return between.join((m.content or "") for m in msgs)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import sys
import urllib.parse

# ensure kani is available in path
sys.path.append("..")

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "kani"
copyright = "2023, Andrew Zhu and Liam Dugan"
author = "Andrew Zhu and Liam Dugan"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx.ext.viewcode",
    "sphinxext.opengraph",  # https://sphinxext-opengraph.readthedocs.io/en/latest/
    "sphinx_inline_tabs",  # https://sphinx-inline-tabs.readthedocs.io/en/latest/usage.html
    "sphinx_copybutton",  # https://sphinx-copybutton.readthedocs.io/en/latest/
    "sphinxemoji.sphinxemoji",  # https://sphinxemojicodes.readthedocs.io/en/stable/
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_extra_path = ["_extra"]
html_logo = "_static/kani-logo@512.png"
html_favicon = "_extra/favicon.ico"

# maximum_signature_line_length = 88
add_module_names = False
nitpicky = True
nitpick_ignore_regex = [
    (r"py:class", r"aiohttp\..*"),  # aiohttp intersphinx is borked
    (r"py:class", r"torch\..*"),  # idk if torch has intersphinx
    (r"py:class", r"kani\.engines\.openai\.models\..*"),  # internal models are just pydantic models of the oai api
]

# sphinx.ext.autodoc
autoclass_content = "both"
autodoc_member_order = "bysource"

# sphinx.ext.intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# sphinx.ext.githubpages
html_baseurl = ""

# sphinxext.opengraph
ogp_social_cards = {
    "enable": False,  # the bundled Roboto font does not support kanji
}

# sphinx_copybutton
copybutton_exclude = ".linenos, .gp, .go"
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_copy_empty_lines = False

# sphinxemoji.sphinxemoji
sphinxemoji_style = "twemoji"



"""Common builder for the LLaMAv2-chat prompt.

This file is responsible for implementing the common non-strict ChatMessage to tokens translation, while handling
the nuance of the INST and SYS tokens as best as possible.
"""
from typing import Callable

from kani.models import ChatMessage, ChatRole

B_INST, E_INST = "[INST]", "[/INST]"
B_SYS, E_SYS = "<<SYS>>\n", "\n<</SYS>>\n\n"


def build(messages: list[ChatMessage], tokenize: Callable[[str], list[int]], eos_token_id: int = 2) -> list[int]:
    tokens = []
    prompt_buf = []  # parts of the user-assistant pair
    for message in messages:
        if message.role == ChatRole.USER:
            prompt_buf.append(f"{B_INST} {message.content} {E_INST}")
        elif message.role == ChatRole.ASSISTANT:
            prompt_buf.append(f" {message.content} ")
            # turn the current round into tokens
            prompt_round = "".join(prompt_buf)
            # hack: if we see a " {E_INST}{B_INST} " we should replace it with empty string
            # (it happens immediately after a system + user message)
            prompt_round.replace(f" {E_INST}{B_INST} ", "")
            tokens.extend(tokenize(prompt_round))  # assumption: tokenize() adds the BOS token but not the EOS token
            tokens.append(eos_token_id)
            prompt_buf.clear()
        else:
            prompt_buf.append(f"{B_INST} {B_SYS}{message.content}{E_SYS} {E_INST}")
    # flush rest of prompt buffer (probably a user message) into tokens
    if prompt_buf:
        tokens.extend(tokenize("".join(prompt_buf)))
    return tokens



import os
import tempfile

import pytest


@pytest.fixture
def gh_log():
    """Get the GitHub Actions step summary file."""
    fp = os.getenv("GITHUB_STEP_SUMMARY")
    if fp is None:
        with tempfile.TemporaryFile(mode="a") as f:
            yield f
    else:
        with open(fp, mode="a") as f:
            yield f



"""LLaMA v2 (https://ai.meta.com/llama/) is a language model released by Meta that has variants fine-tuned for chat.

This example shows how you can use kani to run a language model on your own machine! See the source code of
:class:`.LlamaEngine` for implementation details.
"""

from kani import Kani, chat_in_terminal
from kani.engines.huggingface.llama2 import LlamaEngine

engine = LlamaEngine(use_auth_token=True, strict=True)
ai = Kani(
    engine,
    system_prompt=(
        "You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe."
        " Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content."
        " Please ensure that your responses are socially unbiased and positive in nature.\n\nIf a question does not"
        " make any sense, or is not factually coherent, explain why instead of answering something not correct. If you"
        " don't know the answer to a question, please don't share false information."
    ),
)

if __name__ == "__main__":
    chat_in_terminal(ai)



class KaniException(Exception):
    """Base class for all Kani exceptions/errors."""


class MessageTooLong(KaniException):
    """This chat message will never fit in the context window."""


# ==== HTTP ====
class HTTPException(KaniException):
    """Base class for all HTTP errors (for HTTP engines)."""


class HTTPTimeout(HTTPException):
    """Timeout occurred connecting to or waiting for a response from an HTTP request."""


class HTTPStatusException(HTTPException):
    """The HTTP server returned a non-200 status code."""

    def __init__(self, status_code: int, msg: str):
        super().__init__(msg)
        self.status_code = status_code


# ==== function calling ====
class FunctionCallException(KaniException):
    """Base class for exceptions that occur when a model calls an @ai_function."""

    def __init__(self, retry: bool):
        self.retry = retry


class WrappedCallException(FunctionCallException):
    """The @ai_function raised an exception."""

    def __init__(self, retry, original):
        super().__init__(retry)
        self.original = original

    def __str__(self):
        return str(self.original)


class NoSuchFunction(FunctionCallException):
    """The model attempted to call a function that does not exist."""

    def __init__(self, name):
        super().__init__(True)
        self.name = name


# ==== programmer errors ====
class FunctionSpecError(KaniException):
    """This @ai_function spec is invalid."""


class MissingModelDependencies(KaniException):
    """You are trying to use an engine but do not have engine-specific packages installed."""



import abc

from kani.ai_function import AIFunction
from kani.exceptions import MissingModelDependencies
from kani.models import ChatMessage
from ..base import BaseEngine, Completion

try:
    import torch
    from transformers import AutoModelForCausalLM, AutoTokenizer
except ImportError:
    raise MissingModelDependencies(
        'The HuggingEngine requires extra dependencies. Please install kani with "pip install kani[huggingface]". '
        "You will also need to install PyTorch manually."
    ) from None


class HuggingEngine(BaseEngine, abc.ABC):
    """Base engine for all HuggingFace text-generation models.

    This class implements the main decoding logic for any HuggingFace model based on a pretrained
    ``AutoModelForCausalLM``. To implement a new HuggingFace model, just implement :meth:`build_prompt` and
    :meth:`~.BaseEngine.message_len` for the specified model.

    .. tip:: See :ref:`4b_quant` for information about loading a quantized model for lower memory usage.
    """

    def __init__(
        self,
        model_id: str,
        max_context_size: int,
        use_auth_token=None,
        device: str | None = None,
        tokenizer_kwargs: dict = None,
        model_load_kwargs: dict = None,
        **hyperparams,
    ):
        """
        :param model_id: The ID of the model to load from HuggingFace.
        :param max_context_size: The context size of the model.
        :param use_auth_token: The Hugging Face access token (for gated models). Pass True to load from huggingface-cli.
        :param device: The hardware device to use. If not specified, uses CUDA if available; otherwise uses CPU.
        :param tokenizer_kwargs: Additional arguments to pass to ``AutoTokenizer.from_pretrained()``.
        :param model_load_kwargs: Additional arguments to pass to ``AutoModelForCausalLM.from_pretrained()``.
        :param hyperparams: Additional arguments to supply the model during generation.
        """
        if tokenizer_kwargs is None:
            tokenizer_kwargs = {}
        if model_load_kwargs is None:
            model_load_kwargs = {}

        tokenizer_kwargs.setdefault("use_auth_token", use_auth_token)
        model_load_kwargs.setdefault("use_auth_token", use_auth_token)

        self.model_id = model_id
        self.max_context_size = max_context_size
        self.tokenizer = AutoTokenizer.from_pretrained(model_id, **tokenizer_kwargs)
        self.model = AutoModelForCausalLM.from_pretrained(model_id, **model_load_kwargs)
        self.hyperparams = hyperparams

        if device is None:
            device = "cuda" if torch.has_cuda else "cpu"
        self.device = device
        if self.model.device.type != self.device:
            self.model.to(device)

    @abc.abstractmethod
    def build_prompt(
        self, messages: list[ChatMessage], functions: list[AIFunction] | None = None
    ) -> str | torch.Tensor:
        """Given the list of messages from kani, build either a single string representing the prompt for the model,
        or build the token tensor."""
        raise NotImplementedError

    async def predict(
        self, messages: list[ChatMessage], functions: list[AIFunction] | None = None, **hyperparams
    ) -> Completion:
        """
        Given the current context of messages and available functions, get the next predicted chat message from the LM.

        :param messages: The messages in the current chat context. ``sum(message_len(m) for m in messages)`` is
            guaranteed to be less than max_context_size.
        :param functions: The functions the LM is allowed to call.
        :param hyperparams: Any additional parameters to pass to GenerationMixin.generate(). (See
            https://huggingface.co/docs/transformers/main_classes/text_generation)
        """
        prompt = self.build_prompt(messages, functions)
        if isinstance(prompt, str):
            # prompt str to tokens
            tokenized = self.tokenizer(prompt, return_tensors="pt", return_length=True)
            input_len = int(tokenized.length)
            input_toks = tokenized.input_ids
        elif isinstance(prompt, torch.Tensor):
            input_toks = prompt
            input_len = len(input_toks[0])
        else:
            raise TypeError("build_prompt should either return a str or a Tensor.")
        # move the input tensor to the right device
        if input_toks.device.type != self.device:
            input_toks = input_toks.to(self.device)
        # set up hyperparams for HF decode
        hyperparams = {**self.hyperparams, **hyperparams}
        hyperparams.setdefault("max_length", self.max_context_size)  # by default HF sets this to 20, which is too small
        # run it through the model
        output = self.model.generate(input_toks, **hyperparams)
        # decode to tokens
        # the completion shouldn't include the prompt or stop token
        content = self.tokenizer.decode(output[0][input_len:-1]).strip()
        return Completion(ChatMessage.assistant(content), prompt_tokens=input_len, completion_tokens=len(output[0]) - 1)



"""Example from Advanced Usage docs.

This example shows how kani's function calling functionality can be used to retrieve information from an external
database, e.g. Wikipedia.
"""
import json
import os
from typing import Annotated

from kani import AIParam, Kani, ai_function, chat_in_terminal
from kani.engines.httpclient import BaseClient
from kani.engines.openai import OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")

engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")


# let's define a client using kani's BaseClient:
class WikipediaClient(BaseClient):
    SERVICE_BASE = "https://en.wikipedia.org/w/api.php"


class WikipediaRetrievalKani(Kani):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.wikipedia_client = WikipediaClient()

    @ai_function()
    async def wikipedia(
        self,
        title: Annotated[str, AIParam(desc='The article title on Wikipedia, e.g. "Train_station".')],
    ):
        """Get additional information about a topic from Wikipedia."""
        # https://en.wikipedia.org/w/api.php?action=query&format=json&prop=extracts&titles=Train&explaintext=1&formatversion=2
        resp = await self.wikipedia_client.get(
            "/",
            params={
                "action": "query",
                "format": "json",
                "prop": "extracts",
                "titles": title,
                "explaintext": 1,
                "formatversion": 2,
            },
        )
        page = resp["query"]["pages"][0]
        if extract := page.get("extract"):
            return extract
        return f"The page {title!r} does not exist on Wikipedia."

    @ai_function()
    async def search(self, query: str):
        """Find titles of Wikipedia articles similar to the given query."""
        # https://en.wikipedia.org/w/api.php?action=opensearch&format=json&search=Train
        resp = await self.wikipedia_client.get("/", params={"action": "opensearch", "format": "json", "search": query})
        return json.dumps(resp[1])


ai = WikipediaRetrievalKani(engine)
if __name__ == "__main__":
    chat_in_terminal(ai)



import abc

from kani.ai_function import AIFunction
from kani.exceptions import MissingModelDependencies
from kani.models import ChatMessage
from ..base import BaseEngine, Completion

try:
    from ctransformers import AutoModelForCausalLM
except ImportError as e:
    raise MissingModelDependencies(
        'The CTransformersEngine requires extra dependencies. Please install kani with "pip install'
        ' kani[ctransformers]".'
    ) from None


class CTransformersEngine(BaseEngine, abc.ABC):
    """CTransformers Engine

    This class implements the main decoding logic for any GGML model based on a pretrained ``AutoModelForCausalLM``.
    """

    def __init__(
        self,
        model_id: str,
        model_file: str = None,
        max_context_size: int = 1024,
        model_load_kwargs: dict = None,
        **hyperparams,
    ):
        """
        :param model_id: The ID of the model to load from HuggingFace or locally.
        :param model_file: The file of the model to load from HuggingFace repo or locally.
        :param max_context_size: The context size of the model.
        :param model_load_kwargs: Additional arguments to pass to ``AutoModelForCausalLM.from_pretrained()``.
        :param hyperparams: Additional arguments to supply the model during generation.
        """

        if model_load_kwargs is None:
            model_load_kwargs = {}

        self.model_id = model_id
        self.model_file = model_file
        self.max_context_size = max_context_size
        model_load_kwargs.setdefault("context_length", self.max_context_size)
        self.model = AutoModelForCausalLM.from_pretrained(model_id, model_file=model_file, **model_load_kwargs)
        self.hyperparams = hyperparams

    @abc.abstractmethod
    def build_prompt(self, messages: list[ChatMessage], functions: list[AIFunction] | None = None) -> str | list[int]:
        """Given the list of messages from kani, build either a single string representing the prompt for the model,
        or build the token tensor."""
        raise NotImplementedError

    async def predict(
        self, messages: list[ChatMessage], functions: list[AIFunction] | None = None, **hyperparams
    ) -> Completion:
        """
        Given the current context of messages and available functions, get the next predicted chat message from the LM.

        :param messages: The messages in the current chat context. ``sum(message_len(m) for m in messages)`` is
            guaranteed to be less than max_context_size.
        :param functions: The functions the LM is allowed to call.
        :param hyperparams: Any additional parameters to pass to GenerationMixin.generate(). (See
            https://github.com/marella/ctransformers#method-llmgenerate)
        """
        prompt = self.build_prompt(messages, functions)
        if isinstance(prompt, str):
            # prompt str to tokens
            input_toks = self.model.tokenize(prompt)
            input_len = len(input_toks)
        elif isinstance(prompt, list):
            input_toks = prompt
            input_len = len(input_toks)
        else:
            raise TypeError("build_prompt should either return a str or a list[int].")

        output_toks = list(self.model.generate(input_toks, **hyperparams))
        output_len = len(output_toks)
        content = self.model.detokenize(output_toks).strip()

        return Completion(ChatMessage.assistant(content), prompt_tokens=input_len, completion_tokens=output_len)



import random
import string

from kani import ChatMessage, ChatRole, Kani
from tests.engine import TestEngine
from tests.utils import flatten_chatmessages

engine = TestEngine()


async def test_chat_round():
    ai = Kani(engine, desired_response_tokens=3)
    # 5 tokens, no omitting
    resp = await ai.chat_round("12345")
    assert resp.content == "a"
    assert resp.role == ChatRole.ASSISTANT
    assert len(ai.chat_history) == 2
    prompt = await ai.get_prompt()
    assert len(prompt) == 2  # 12345a
    assert flatten_chatmessages(prompt) == "12345a"

    # 5 more, first message should be omitted
    resp = await ai.chat_round_str("67890")
    assert resp == "a"
    assert len(ai.chat_history) == 4
    assert flatten_chatmessages(ai.chat_history) == "12345a67890a"
    prompt = await ai.get_prompt()
    assert len(prompt) == 3  # 12345a
    assert flatten_chatmessages(prompt) == "a67890a"


async def test_always_include():
    # always include 2 tokens, reserve 3 for response
    ai = Kani(engine, desired_response_tokens=3, system_prompt="1", always_included_messages=[ChatMessage.user("2")])
    assert len(ai.always_included_messages) == 2
    assert sum(ai.message_token_len(m) for m in ai.always_included_messages) == 2

    # messages are only included if <= 5 tokens
    resp = await ai.chat_round_str("12345")
    assert resp == "a"
    assert len(ai.chat_history) == 2  # always include are not included in chat history
    assert flatten_chatmessages(ai.chat_history) == "12345a"
    prompt = await ai.get_prompt()
    assert len(prompt) == 3  # 12a (12345 gets dropped)
    assert flatten_chatmessages(prompt) == "12a"


async def test_spam():
    # spam the kani with a bunch of random prompts
    # and make sure it never breaks
    ai = Kani(engine, desired_response_tokens=3, system_prompt="1", always_included_messages=[ChatMessage.user("2")])
    for _ in range(1000):
        query_len = random.randint(0, 5)
        query = "".join(random.choice(string.ascii_letters) for _ in range(query_len))
        resp = await ai.chat_round_str(query, test_echo=True)
        assert resp == query

        prompt = await ai.get_prompt()
        assert sum(ai.message_token_len(m) for m in prompt) <= ai.max_context_size



import enum
import os
from typing import Annotated

from kani import AIParam, ChatMessage, FunctionCall, Kani, ai_function, chat_in_terminal
from kani.engines.openai import OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")


# from 2_function_calling_weather - skip down to L33!
class Unit(enum.Enum):
    FAHRENHEIT = "fahrenheit"
    CELSIUS = "celsius"


class MyKani(Kani):
    @ai_function()
    def get_weather(
        self,
        location: Annotated[str, AIParam(desc="The city and state, e.g. San Francisco, CA")],
        unit: Unit,
    ):
        """Get the current weather in a given location."""
        # call some weather API, or just mock it for this example
        degrees = 72 if unit == Unit.FAHRENHEIT else 22
        return f"Weather in {location}: Sunny, {degrees} degrees {unit.value}."


# end 2_function_calling_weather

# build the chat history with examples
fewshot = [
    ChatMessage.user("What's the weather in Philadelphia?"),
    ChatMessage.assistant(
        content=None,
        function_call=FunctionCall.with_args("get_weather", location="Philadelphia, PA", unit="fahrenheit"),
    ),
    ChatMessage.function("get_weather", "Weather in Philadelphia, PA: Partly cloudy, 85 degrees fahrenheit."),
    ChatMessage.assistant(
        content=None,
        function_call=FunctionCall.with_args("get_weather", location="Philadelphia, PA", unit="celsius"),
    ),
    ChatMessage.function("get_weather", "Weather in Philadelphia, PA: Partly cloudy, 29 degrees celsius."),
    ChatMessage.assistant("It's currently 85F (29C) and partly cloudy in Philadelphia."),
]
# and give it to the kani when you initialize it
ai = MyKani(engine, chat_history=fewshot)

if __name__ == "__main__":
    chat_in_terminal(ai)



"""
As OpenAI doesn't tell us exactly how functions are exposed to GPT, we have to rely on some community reverse
engineering to build a reliable method to reserve tokens for AI Functions.

See https://community.openai.com/t/how-to-calculate-the-tokens-when-using-function-call/266573 for more details.
"""
import json
import warnings

from kani.ai_function import AIFunction


def prompt(functions: list[AIFunction]) -> str:
    return "".join(map(format_function, functions))


def format_function(function: AIFunction) -> str:
    # Thanks @CGamesPlay for https://gist.github.com/CGamesPlay/dd4f108f27e2eec145eedf5c717318f5, which this
    # implementation is based on.
    def resolve_ref(schema):
        if schema.get("$ref") is not None:
            *_, ref = schema["$ref"].rsplit("/", 1)
            schema = json_schema["$defs"][ref]
        return schema

    def format_schema(schema, indent):
        schema = resolve_ref(schema)
        if "enum" in schema:
            return format_enum(schema)
        elif schema["type"] == "object":
            return format_object(schema, indent)
        elif schema["type"] == "array":
            return format_schema(schema["items"], indent) + "[]"
        elif schema["type"] in ("string", "number", "integer", "boolean", "null"):  # these are all 1 token!
            return schema["type"]
        warnings.warn(
            f"Unknown JSON schema type estimating tokens for OpenAI: {schema['type']!r}\n"
            "The returned estimate may be off by a significant amount."
        )
        return schema["type"]

    def format_enum(schema):
        return " | ".join(json.dumps(o) for o in schema["enum"])

    def format_object(schema, indent):
        result = "{\n"
        if "properties" not in schema or len(schema["properties"]) == 0:
            if schema.get("additionalProperties", False):
                return "object"
            return None
        for key, value in schema["properties"].items():
            value = resolve_ref(value)
            value_rendered = format_schema(value, indent + 1)
            if value_rendered is None:
                continue
            if "description" in value and indent == 0:
                for line in value["description"].strip().split("\n"):
                    result += f"{'  ' * indent}// {line}\n"
            optional = "" if key in schema.get("required", {}) else "?"
            comment = "" if value.get("default") is None else f" // default: {format_default(value)}"
            result += f"{'  ' * indent}{key}{optional}: {value_rendered},{comment}\n"
        result += ("  " * (indent - 1)) + "}"
        return result

    def format_default(schema):
        v = schema["default"]
        if schema["type"] == "number":
            return f"{v:.1f}" if float(v).is_integer() else str(v)
        else:
            return str(v)

    json_schema = function.json_schema
    if function.desc:
        out = f"// {function.desc}\ntype {function.name} = ("
    else:
        out = f"type {function.name} = ("
    formatted = format_object(json_schema, 0)
    if formatted is not None:
        out += "_: " + formatted
    out += ") => any;\n\n"
    return out



from kani.ai_function import AIFunction
from kani.models import ChatMessage, ChatRole
from .. import llama2_prompt
from .base import CTransformersEngine


class LlamaCTransformersEngine(CTransformersEngine):
    """Implementation of LLaMA v2 using ctransformers.

    You may also use the 7b or 13b models that use the LLaMA prompt.
    These models are converted from the original models in
    pytorch format to GGML format and can be found on huggingface.
    You can also download these models locally and pass the local path as an argument to the class.

    The huggingface GGML repos generally have multiple models in them (of different quantization levels),
    so you can choose the model depending inference speed, memory, and quality tradeoffs depending on the quantization
    performed on the model.

    A specific GGML model can be used from huggingface by passing the model id and the model file
    (e.g. https://huggingface.co/TheBloke/Llama-2-7B-Chat-GGML#provided-files) to
    `LlamaCTransformersEngine` and the model will be automatically downloaded and placed locally.

    Model IDs:

    - ``TheBloke/Llama-2-7B-Chat-GGML``
    - ``TheBloke/Llama-2-13B-chat-GGML``

    .. code-block:: python

        engine = LlamaCTransformersEngine("TheBloke/Llama-2-7B-Chat-GGML", model_file="llama-2-7b-chat.ggmlv3.q5_K_M.bin")
        ai = Kani(engine)

    .. tip:: The ``LlamaCTransformersEngine`` is a drop-in replacement for the :class:`.LlamaEngine`.

    .. seealso:: https://github.com/marella/ctransformers/tree/main#method-llmgenerate for generation parameters
    """

    def __init__(
        self,
        model_id: str = "TheBloke/Llama-2-7B-Chat-GGML",
        model_file: str = "llama-2-7b-chat.ggmlv3.q5_K_M.bin",
        *args,
        **kwargs,
    ):
        """
        :param model_id: The ID of the model to load from HuggingFace.
        :param model_file: The file of the model to load from HuggingFace repo or locally.
        :param max_context_size: The context size of the model.
        :param model_load_kwargs: Additional arguments to pass to ``AutoModelForCausalLM.from_pretrained()``.
        :param hyperparams: Additional arguments to supply the model during generation.
        """

        kwargs.setdefault("max_context_size", 4096)  # LLaMA has 4096 token window
        super().__init__(model_id, model_file, *args, **kwargs)

    def build_prompt(self, messages: list[ChatMessage], functions: list[AIFunction] | None = None) -> list[int]:
        return llama2_prompt.build(messages, tokenize=self.model.tokenize, eos_token_id=self.model.eos_token_id)

    def message_len(self, message: ChatMessage) -> int:
        # https://github.com/facebookresearch/llama/blob/main/llama/generation.py#L212
        if message.role == ChatRole.USER:
            # <s> [INST] {} [/INST] -> 7
            return len(self.model.tokenize(message.content)) + 7
        elif message.role == ChatRole.ASSISTANT:
            # {} </s> -> 2
            return len(self.model.tokenize(f" {message.content} ")) + 2
        # <s> [INST] <<SYS>>\n{}\n<</SYS>>\n\n [/INST] -> 20
        return len(self.model.tokenize(message.content)) + 20



"""Vicuna (https://huggingface.co/lmsys/vicuna-7b-v1.3) is a language model based on LLaMA that is fine-tuned for chat.

This example shows how you can use kani to run a language model on your own machine! See the source code of
:class:`.VicunaEngine` for implementation details.
"""

from kani import Kani, chat_in_terminal
from kani.engines.huggingface.vicuna import VicunaEngine

engine = VicunaEngine()
ai = Kani(
    engine,
    system_prompt=(
        "A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed,"
        " and polite answers to the user's questions."
    ),
)
if __name__ == "__main__":
    chat_in_terminal(ai)



from kani.ai_function import AIFunction
from kani.exceptions import MissingModelDependencies
from kani.models import ChatMessage, ChatRole
from .. import llama2_prompt
from ..llama2_prompt import B_INST, B_SYS, E_INST, E_SYS
from .base import HuggingEngine

try:
    import sentencepiece
    import torch
    from torch import tensor
except ImportError:
    raise MissingModelDependencies(
        'The LlamaEngine requires extra dependencies. Please install kani with "pip install'
        " 'kani[huggingface,llama]'\". You will also need to install PyTorch manually."
    ) from None


class LlamaEngine(HuggingEngine):
    """Implementation of LLaMA v2 using huggingface transformers.

    You may also use the 13b, 70b, or other LLaMA models that use the LLaMA prompt by passing the HuggingFace model
    ID to the initializer.

    Model IDs:

    - ``meta-llama/Llama-2-7b-chat-hf``
    - ``meta-llama/Llama-2-13b-chat-hf``
    - ``meta-llama/Llama-2-70b-chat-hf``

    In theory you could also use the non-chat-tuned variants as well.

    .. attention::

        You will need to accept Meta's license in order to download the LLaMA v2 weights. Visit
        https://ai.meta.com/resources/models-and-libraries/llama-downloads/ and
        https://huggingface.co/meta-llama/Llama-2-7b-chat-hf to request access.

        Then, run ``huggingface-cli login`` to authenticate with Hugging Face.

    .. seealso:: https://huggingface.co/meta-llama/Llama-2-7b-chat-hf

    .. tip:: See :ref:`4b_quant` for information about loading a quantized model for lower memory usage.

    .. code-block:: python

        engine = LlamaEngine("meta-llama/Llama-2-7b-chat-hf", use_auth_token=True, strict=True)
        ai = Kani(engine)
    """

    def __init__(self, model_id: str = "meta-llama/Llama-2-7b-chat-hf", *args, strict=False, **kwargs):
        """
        :param model_id: The ID of the model to load from HuggingFace.
        :param use_auth_token: The Hugging Face access token (for gated models). Pass True to load from huggingface-cli.
        :param strict: Whether to enforce the same prompt constraints as Meta's LLaMA (always starting with
            system, alternating user/assistant). This is recommended for using the base LLaMA model.
        :param max_context_size: The context size of the model.
        :param device: The hardware device to use. If not specified, uses CUDA if available; otherwise uses CPU.
        :param tokenizer_kwargs: Additional arguments to pass to ``AutoTokenizer.from_pretrained()``.
        :param model_load_kwargs: Additional arguments to pass to ``AutoModelForCausalLM.from_pretrained()``.
        :param hyperparams: Additional arguments to supply the model during generation.
        """
        self.strict = strict

        kwargs.setdefault("max_context_size", 4096)  # LLaMA has 4096 token window
        super().__init__(model_id, *args, **kwargs)

    # noinspection PyMethodMayBeStatic
    def _build_prompt_strict(self, messages: list[ChatMessage]):
        if len(messages) < 2:
            raise ValueError("You must specify at least one message and a system prompt (LLaMA strict mode).")
        if messages[0].role != ChatRole.SYSTEM:
            raise ValueError("The first message must be a system prompt (LLaMA strict mode).")
        if not (
            all([m.role == ChatRole.USER for m in messages[1::2]])
            and all([m.role == ChatRole.ASSISTANT for m in messages[2::2]])
        ):
            raise ValueError("Messages after the first must alternate between user and system (LLaMA strict mode).")
        if messages[-1].role != ChatRole.USER:
            raise ValueError("The last message must be from the user (LLaMA strict mode).")
        # implementation based on llama code
        dialog = [f"{B_SYS}{messages[0].content}{E_SYS}{messages[1].content}"] + [m.content for m in messages[2:]]
        dialog_tokens = sum(
            [
                self.tokenizer.encode(f"{B_INST} {prompt} {E_INST} {answer}") + [self.tokenizer.eos_token_id]
                for prompt, answer in zip(dialog[::2], dialog[1::2])
            ],
            [],
        )
        dialog_tokens += self.tokenizer.encode(f"{B_INST} {dialog[-1]} {E_INST}")
        return torch.tensor([dialog_tokens], device=self.device)

    def build_prompt(self, messages: list[ChatMessage], functions: list[AIFunction] | None = None) -> torch.Tensor:
        if self.strict:
            return self._build_prompt_strict(messages)
        # non-strict has to kind of just do its best
        # ganbatte, kani, ganbatte
        tokens = llama2_prompt.build(messages, tokenize=self.tokenizer.encode, eos_token_id=self.tokenizer.eos_token_id)
        return torch.tensor([tokens], device=self.device)

    def message_len(self, message: ChatMessage) -> int:
        # https://github.com/facebookresearch/llama/blob/main/llama/generation.py#L212
        if message.role == ChatRole.USER:
            # <s> [INST] {} [/INST] -> 7
            return self.tokenizer(message.content, return_length=True).length[0] + 7
        elif message.role == ChatRole.ASSISTANT:
            # {} </s> -> 2
            return self.tokenizer(f" {message.content} ", return_length=True).length[0] + 2
        # <s> [INST] <<SYS>>\n{}\n<</SYS>>\n\n [/INST] -> 20
        return self.tokenizer(message.content, return_length=True).length[0] + 20



import asyncio
import functools
import inspect
import typing
from typing import Annotated

from pydantic import validate_call

from .exceptions import FunctionSpecError
from .json_schema import AIParamSchema, create_json_schema
from .models import ChatRole


class AIFunction:
    """Wrapper around a function to expose to a language model."""

    def __init__(
        self,
        inner,
        after: ChatRole = ChatRole.ASSISTANT,
        name: str | None = None,
        desc: str | None = None,
        auto_retry: bool = True,
        json_schema: dict | None = None,
        auto_truncate: int | None = None,
    ):
        """
        :param inner: The function implementation.
        :param after: Who should speak next after the function call completes (see :ref:`next_actor`). Defaults to the
            model.
        :param name: The name of the function (defaults to the name of the function in source code).
        :param desc: The function's description (defaults to the function's docstring).
        :param auto_retry: Whether the model should retry calling the function if it gets it wrong
            (see :ref:`auto_retry`).
        :param json_schema: A JSON Schema document describing the function's parameters. By default, kani will
            automatically generate one, but this can be helpful for overriding it in any tricky cases.
        :param auto_truncate: If a function response is longer than this many tokens, truncate it until it is at most
            this many tokens and add "..." to the end. By default, no responses will be truncated. This uses a smart
            paragraph-aware truncation algorithm.
        """
        # pydantic's wrapper mangles the async signature so we have to store this here
        self._inner_is_coro = inspect.iscoroutinefunction(inner)
        self.inner = validate_call(inner)
        self.after = after
        self.name = name or inner.__name__
        self.desc = desc or inspect.getdoc(inner)
        self.auto_retry = auto_retry
        self.json_schema = self.create_json_schema() if json_schema is None else json_schema
        self.auto_truncate = auto_truncate

        # wraps() things
        self.__name__ = inner.__name__
        self.__qualname__ = inner.__qualname__
        self.__annotations__ = inner.__annotations__
        self.__module__ = inner.__module__
        self.__doc__ = inner.__doc__

    async def __call__(self, *args, **kwargs):
        if self._inner_is_coro:
            return await self.inner(*args, **kwargs)
        # run synch functions in a threadpool in order to maintain async safety
        inner_partial = functools.partial(self.inner, *args, **kwargs)
        return await asyncio.get_event_loop().run_in_executor(None, inner_partial)

    def create_json_schema(self) -> dict:
        """Create a JSON schema representing this function's parameters as a JSON object."""
        # get list of params
        params = []
        sig = inspect.signature(self.inner)
        type_hints = typing.get_type_hints(self.inner)
        for name, param in sig.parameters.items():
            # ensure param can be supplied thru kwargs
            if param.kind not in (inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY):
                raise FunctionSpecError(
                    "Positional-only or variadic parameters are not allowed in @ai_function()s."
                    f" ({self.inner.__name__}#{name})"
                )

            # ensure the type is annotated
            annotation = param.annotation
            if annotation is inspect.Parameter.empty:
                raise FunctionSpecError(
                    f"All @ai_function() parameters must have a type annotation ({self.inner.__name__}#{name})."
                )

            # ensure type hint matches up
            if name not in type_hints:
                raise RuntimeError(f"The schema generator could not find the type hint ({self.inner.__name__}#{name}).")

            # get aiparam and add it to the list
            ai_param = get_aiparam(annotation)
            params.append(AIParamSchema(name=name, t=type_hints[name], default=param.default, aiparam=ai_param))
        # create a schema generator and generate
        return create_json_schema(params)


def ai_function(
    func=None,
    *,
    after: ChatRole = ChatRole.ASSISTANT,
    name: str | None = None,
    desc: str | None = None,
    auto_retry: bool = True,
    json_schema: dict | None = None,
    auto_truncate: int | None = None,
):
    """Decorator to mark a method of a Kani to expose to the AI.

    :param after: Who should speak next after the function call completes (see :ref:`next_actor`). Defaults to the
        model.
    :param name: The name of the function (defaults to the name of the function in source code).
    :param desc: The function's description (defaults to the function's docstring).
    :param auto_retry: Whether the model should retry calling the function if it gets it wrong (see :ref:`auto_retry`).
    :param json_schema: A JSON Schema document describing the function's parameters. By default, kani will automatically
        generate one, but this can be helpful for overriding it in any tricky cases.
    :param auto_truncate: If a function response is longer than this many tokens, truncate it until it is at most
        this many tokens and add "..." to the end. By default, no responses will be truncated. This uses a smart
        paragraph-aware truncation algorithm.
    """

    def deco(f):
        f.__ai_function__ = {
            "after": after,
            "name": name or f.__name__,
            "desc": desc or inspect.getdoc(f),
            "auto_retry": auto_retry,
            "json_schema": json_schema,
            "auto_truncate": auto_truncate,
        }
        return f

    if func is not None:
        return deco(func)
    return deco


# ==== AIParam ====
class AIParam:
    """Special tag to annotate types with in order to provide parameter-level metadata to kani."""

    def __init__(self, desc: str):
        self.desc = desc


def get_aiparam(annotation: type) -> AIParam | None:
    """If the type annotation is an Annotated containing an AIParam, extract and return it."""
    # is it Annotated? if so, get AIParam from annotation
    if typing.get_origin(annotation) is not Annotated:
        return

    for a in typing.get_args(annotation):
        if isinstance(a, AIParam):
            return a



import asyncio
from typing import Literal, overload

import aiohttp
import pydantic

from kani.models import ChatMessage
from ..httpclient import BaseClient, HTTPException, HTTPStatusException, HTTPTimeout
from .models import ChatCompletion, Completion, FunctionSpec, SpecificFunctionCall


class OpenAIClient(BaseClient):
    """Simple HTTP client to interface with the OpenAI API."""

    def __init__(
        self,
        api_key: str,
        http: aiohttp.ClientSession = None,
        organization: str = None,
        retry: int = 5,
        api_base: str = "https://api.openai.com/v1",
        headers: dict = None,
    ):
        if headers is None:
            headers = {}
        super().__init__(http)
        self.api_key = api_key
        self.organization = organization
        self.retry = retry
        self.SERVICE_BASE = api_base
        self.headers = headers

    async def request(self, method: str, route: str, headers=None, retry=None, **kwargs):
        if headers is None:
            headers = self.headers.copy()
        # set up auth headers
        if "Authorization" not in headers:
            headers["Authorization"] = f"Bearer {self.api_key}"
        if self.organization:
            headers["OpenAI-Organization"] = self.organization

        # make the request
        retry = retry if retry is not None else self.retry
        for i in range(retry):
            try:
                return await super().request(method, route, headers=headers, **kwargs)
            except (HTTPStatusException, HTTPTimeout) as e:
                if (i + 1) == retry:
                    raise
                retry_sec = 2**i
                self.logger.warning(f"OpenAI returned {e}, retrying in {retry_sec} sec...")
                await asyncio.sleep(retry_sec)
        raise RuntimeError("ran out of retries but no error encountered, halp")

    # ==== completions ====
    @overload
    async def create_completion(
        self,
        model: str,
        prompt: str = "<|endoftext|>",
        suffix: str = None,
        max_tokens: int = 16,
        temperature: float = 1.0,
        top_p: float = 1.0,
        n: int = 1,
        logprobs: int = None,
        echo: bool = False,
        stop: str | list[str] = None,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        best_of: int = 1,
        logit_bias: dict = None,
        user: str = None,
    ) -> Completion:
        ...

    async def create_completion(self, model: str, **kwargs) -> Completion:
        data = await self.post("/completions", json={"model": model, **kwargs})
        try:
            return Completion.model_validate(data)
        except pydantic.ValidationError as e:
            self.logger.exception(f"Failed to deserialize OpenAI response: {data}")
            raise HTTPException(f"Could not deserialize response: {data}") from e

    # ==== chat ====
    @overload
    async def create_chat_completion(
        self,
        model: str,
        messages: list[ChatMessage],
        functions: list[FunctionSpec] | None = None,
        function_call: SpecificFunctionCall | Literal["auto"] | Literal["none"] | None = None,
        temperature: float = 1.0,
        top_p: float = 1.0,
        n: int = 1,
        stop: str | list[str] | None = None,
        max_tokens: int | None = None,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        logit_bias: dict | None = None,
        user: str | None = None,
    ) -> ChatCompletion:
        ...

    async def create_chat_completion(
        self,
        model: str,
        messages: list[ChatMessage],
        functions: list[FunctionSpec] | None = None,
        **kwargs,
    ) -> ChatCompletion:
        """Create a chat completion.

        See https://platform.openai.com/docs/api-reference/chat/create.
        """
        # transform pydantic models
        if functions:
            kwargs["functions"] = [f.model_dump(exclude_unset=True) for f in functions]
        if "function_call" in kwargs and isinstance(kwargs["function_call"], SpecificFunctionCall):
            kwargs["function_call"] = kwargs["function_call"].model_dump(exclude_unset=True)
        # call API
        data = await self.post(
            "/chat/completions",
            json={
                "model": model,
                "messages": [cm.model_dump(exclude_unset=True, mode="json") for cm in messages],
                **kwargs,
            },
        )
        try:
            return ChatCompletion.model_validate(data)
        except pydantic.ValidationError:
            self.logger.exception(f"Failed to deserialize OpenAI response: {data}")
            raise HTTPException(f"Could not deserialize response: {data}")



from kani import ChatMessage
from kani.engines.base import BaseEngine, Completion


class TestEngine(BaseEngine):
    """A mock engine used for testing.

    Each message has a token length equal to its str length, and predict always returns a one-token message.
    """

    max_context_size = 10

    def message_len(self, message: ChatMessage) -> int:
        return len(message.content)

    async def predict(self, messages, functions=None, test_echo=False, **hyperparams) -> Completion:
        """
        :param test_echo: If True, the prediction echoes the last message.
        """
        assert sum(len(m.content or "") for m in messages) <= self.max_context_size
        if test_echo:
            content = messages[-1].content
            return Completion(ChatMessage.assistant(content))
        return Completion(ChatMessage.assistant("a"))



import asyncio
import inspect
import logging
import weakref
from typing import AsyncIterable, Callable

from .ai_function import AIFunction
from .engines.base import BaseCompletion, BaseEngine
from .exceptions import FunctionCallException, MessageTooLong, NoSuchFunction, WrappedCallException
from .models import ChatMessage, ChatRole, FunctionCall
from .utils.typing import PathLike, SavedKani

log = logging.getLogger("kani")
message_log = logging.getLogger("kani.messages")


class Kani:
    """Base class for all kani.

    **Entrypoints**

    ``chat_round(query: str, **kwargs) -> ChatMessage``

    ``full_round(query: str, function_call_formatter: Callable[[ChatMessage], str], **kwargs) -> AsyncIterable[ChatMessage]``

    ``chat_round_str(query: str, **kwargs) -> str``

    ``full_round_str(query: str, function_call_formatter: Callable[[ChatMessage], str], **kwargs) -> AsyncIterable[str]``

    **Function Calling**

    Subclass and use ``@ai_function()`` to register functions. The schema will be autogenerated from the function
    signature (see :func:`ai_function`).

    To perform a chat round with functions, use :meth:`full_round()` as an async iterator::

        async for msg in kani.full_round(prompt):
            # responses...

    Each response will be a :class:`.ChatMessage`.

    Alternatively, you can use :meth:`full_round_str` and control the format of a yielded function call with
    ``function_call_formatter``.

    **Retry & Model Feedback**

    If the model makes an error when attempting to call a function (e.g. calling a function that does not exist or
    passing params with incorrect and non-coercible types) or the function raises an exception, Kani will send the
    error in a system message to the model, allowing it up to *retry_attempts* to correct itself and retry the call.
    """

    def __init__(
        self,
        engine: BaseEngine,
        system_prompt: str = None,
        always_included_messages: list[ChatMessage] = None,
        desired_response_tokens: int = 450,
        chat_history: list[ChatMessage] = None,
        functions: list[AIFunction] = None,
        retry_attempts: int = 1,
    ):
        """
        :param engine: The LM engine implementation to use.
        :param system_prompt: The system prompt to provide to the LM. The prompt will not be included in
            :attr:`chat_history`.
        :param always_included_messages: A list of messages to always include as a prefix in all chat rounds (i.e.,
            evict newer messages rather than these to manage context length). These will not be included in
            :attr:`chat_history`.
        :param desired_response_tokens: The minimum amount of space to leave in ``max context size - tokens in prompt``.
            To control the maximum number of tokens generated more precisely, you may be able to configure the engine
            (e.g. ``OpenAIEngine(..., max_tokens=250)``).
        :param chat_history: The chat history to start with (not including system prompt or always included messages),
            for advanced use cases. By default, each kani starts with a new conversation session.

            .. caution::
                If you pass another kani's chat history here without copying it, the same list will be mutated!
                Use ``chat_history=mykani.chat_history.copy()`` to pass a copy.
        :param functions: A list of :class:`.AIFunction` to expose to the model (for dynamic function calling).
            Use :func:`.ai_function` to define static functions (see :doc:`function_calling`).
        :param retry_attempts: How many attempts the LM may take if a function call raises an exception.
        """
        self.engine = engine
        self.system_prompt = system_prompt.strip() if system_prompt else None
        self.desired_response_tokens = desired_response_tokens
        self.max_context_size = engine.max_context_size

        self.always_included_messages: list[ChatMessage] = (
            [ChatMessage.system(self.system_prompt)] if system_prompt else []
        ) + (always_included_messages or [])
        """Chat messages that are always included as a prefix in the model's prompt.
        Includes the system message, if supplied."""

        self.chat_history: list[ChatMessage] = chat_history or []
        """All messages in the current chat state, not including system or always included messages."""

        # async to prevent generating multiple responses missing context
        self.lock = asyncio.Lock()

        # function calling
        self.retry_attempts = retry_attempts

        # find all registered ai_functions and save them
        if functions is None:
            functions = []
        self.functions: dict[str, AIFunction] = {f.name: f for f in functions}
        for name, member in inspect.getmembers(self, predicate=inspect.ismethod):
            if not hasattr(member, "__ai_function__"):
                continue
            f: AIFunction = AIFunction(member, **member.__ai_function__)
            if f.name in self.functions:
                raise ValueError(f"AIFunction {f.name!r} is already registered!")
            self.functions[f.name] = f

        # cache
        self._message_tokens = weakref.WeakKeyDictionary()

    # === main entrypoints ===
    async def chat_round(self, query: str, **kwargs) -> ChatMessage:
        """Perform a single chat round (user -> model -> user, no functions allowed).

        This is slightly faster when you are chatting with a kani with no AI functions defined.

        :param query: The contents of the user's chat message.
        :param kwargs: Additional arguments to pass to the model engine (e.g. hyperparameters).
        :returns: The model's reply.
        """
        async with self.lock:
            # add the user's chat input to the state
            self.chat_history.append(ChatMessage.user(query.strip()))

            # and get a completion
            completion = await self.get_model_completion(include_functions=False, **kwargs)
            message = completion.message
            self.chat_history.append(message)
            return message

    async def chat_round_str(self, query: str, **kwargs) -> str:
        """Like :meth:`chat_round`, but only returns the content of the message."""
        msg = await self.chat_round(query, **kwargs)
        return msg.content

    async def full_round(self, query: str, **kwargs) -> AsyncIterable[ChatMessage]:
        """Perform a full chat round (user -> model [-> function -> model -> ...] -> user).

        Yields each of the model's ChatMessages. A ChatMessage must have at least one of (content, function_call).

        Use this in an async for loop, like so::

            async for msg in kani.full_round("How's the weather?"):
                print(msg.content)

        :param query: The content of the user's chat message.
        :param kwargs: Additional arguments to pass to the model engine (e.g. hyperparameters).
        """
        retry = 0
        is_model_turn = True
        async with self.lock:
            self.chat_history.append(ChatMessage.user(query.strip()))

            while is_model_turn:
                # do the model prediction
                completion = await self.get_model_completion(**kwargs)
                message = completion.message
                self.chat_history.append(message)
                yield message

                # if function call, do it and attempt retry if it's wrong
                if not message.function_call:
                    return

                try:
                    is_model_turn = await self.do_function_call(message.function_call)
                except FunctionCallException as e:
                    should_retry = await self.handle_function_call_exception(message.function_call, e, retry)
                    # retry if we have retry attempts left
                    retry += 1
                    if not should_retry:
                        # disable function calling on the next go
                        kwargs = {**kwargs, "include_functions": False}
                    continue
                else:
                    retry = 0

    async def full_round_str(
        self,
        query: str,
        function_call_formatter: Callable[[ChatMessage], str | None] = lambda _: None,
        **kwargs,
    ) -> AsyncIterable[str]:
        """Like :meth:`full_round`, but each yielded element is a str rather than a ChatMessage.

        :param query: The content of the user's chat message.
        :param function_call_formatter: A function that returns a string to yield when the model decides to call a
            function (or None to yield nothing). By default, ``full_round_str`` does not yield on a function call.
        :param kwargs: Additional arguments to pass to the model engine (e.g. hyperparameters).
        """
        async for message in self.full_round(query, **kwargs):
            if text := message.content:
                yield text

            if message.function_call and (fn_msg := function_call_formatter(message)):
                yield fn_msg

    # ==== helpers ====
    @property
    def always_len(self) -> int:
        """Returns the number of tokens that will always be reserved.

        (e.g. for system prompts, always included messages, the engine, and the response).
        """
        return (
            sum(self.message_token_len(m) for m in self.always_included_messages)
            + self.engine.token_reserve
            + self.engine.function_token_reserve(list(self.functions.values()))
            + self.desired_response_tokens
        )

    def message_token_len(self, message: ChatMessage):
        """Returns the number of tokens used by a given message."""
        try:
            return self._message_tokens[message]
        except KeyError:
            mlen = self.engine.message_len(message)
            self._message_tokens[message] = mlen
            return mlen

    async def get_model_completion(self, include_functions: bool = True, **kwargs) -> BaseCompletion:
        """Get the model's completion with the current chat state.

        Compared to :meth:`chat_round` and :meth:`full_round`, this lower-level method does not save the model's reply
        to the chat history or mutate the chat state; it is intended to help with logging or to repeat a call multiple
        times.

        :param include_functions: Whether to pass this kani's function definitions to the engine.
        :param kwargs: Arguments to pass to the model engine.
        """
        # get the current chat state
        messages = await self.get_prompt()
        # log it (message_log includes the number of messages sent and the last message)
        n_messages = len(messages)
        if n_messages == 0:
            message_log.debug("[0]>>> [requested completion with no prompt]")
        else:
            message_log.debug(f"[{n_messages}]>>> {messages[-1]}")

        # get the model's completion at the given state
        if include_functions:
            completion = await self.engine.predict(messages=messages, functions=list(self.functions.values()), **kwargs)
        else:
            completion = await self.engine.predict(messages=messages, **kwargs)

        # cache its length (if the completion isn't saved to state, this weakrefs and gc's later)
        message = completion.message
        self._message_tokens[message] = completion.completion_tokens or self.message_token_len(message)
        # and log it too
        message_log.debug(f"<<< {message}")
        return completion

    # ==== overridable methods ====
    async def get_prompt(self) -> list[ChatMessage]:
        """
        Called each time before asking the LM engine for a completion to generate the chat prompt.
        Returns a list of messages such that the total token count in the messages is less than
        ``(self.max_context_size - self.desired_response_tokens)``.

        Always includes the system prompt plus any always_included_messages at the start of the prompt.

        You may override this to get more fine-grained control over what is exposed in the model's memory at any given
        call.
        """
        always_len = self.always_len
        remaining = max_size = self.max_context_size - always_len
        total_tokens = 0
        to_keep = 0  # messages to keep from the end of chat history
        for message in reversed(self.chat_history):
            # get and check the message's length
            message_len = self.message_token_len(message)
            if message_len > max_size:
                func_help = (
                    ""
                    if message.role != ChatRole.FUNCTION
                    else "You may set `auto_truncate` in the @ai_function to automatically truncate long responses.\n"
                )
                raise MessageTooLong(
                    "The chat message's size is longer than the allowed context window (after including system"
                    " messages, always included messages, and desired response tokens).\n"
                    f"{func_help}Content: {message.content[:100]}..."
                )
            # see if we can include it
            remaining -= message_len
            if remaining >= 0:
                total_tokens += message_len
                to_keep += 1
            else:
                break
        log.debug(
            f"get_prompt() returned {always_len + total_tokens} tokens ({always_len} always) in"
            f" {len(self.always_included_messages) + to_keep} messages"
            f" ({len(self.always_included_messages)} always)"
        )
        if not to_keep:
            return self.always_included_messages
        return self.always_included_messages + self.chat_history[-to_keep:]

    async def do_function_call(self, call: FunctionCall) -> bool:
        """Resolve a single function call.

        By default, any exception raised from this method will be an instance of a :class:`.FunctionCallException`.

        You may implement an override to add instrumentation around function calls (e.g. tracking success counts
        for varying prompts). See :ref:`do_function_call`.

        :returns: True (default) if the model should immediately react; False if the user speaks next.
        :raises NoSuchFunction: The requested function does not exist.
        :raises WrappedCallException: The function raised an exception.
        """
        log.debug(f"Model requested call to {call.name} with data: {call.arguments!r}")
        # get func
        f = self.functions.get(call.name)
        if not f:
            raise NoSuchFunction(call.name)
        # call it
        try:
            result = await f(**call.kwargs)
            result_str = str(result)
            log.debug(f"{f.name} responded with data: {result_str!r}")
        except Exception as e:
            raise WrappedCallException(f.auto_retry, e) from e
        msg = ChatMessage.function(f.name, result_str)
        # if we are auto truncating, check and see if we need to
        if f.auto_truncate is not None:
            message_len = self.message_token_len(msg)
            if message_len > f.auto_truncate:
                log.warning(
                    f"The content returned by {f.name} is too long ({message_len} > {f.auto_truncate} tokens), auto"
                    " truncating..."
                )
                msg = self._auto_truncate_message(msg, max_len=f.auto_truncate)
                log.debug(f"Auto truncate returned {self.message_token_len(msg)} tokens.")
        # save the result to the chat history
        self.chat_history.append(msg)
        # yield whose turn it is
        return f.after == ChatRole.ASSISTANT

    async def handle_function_call_exception(
        self, call: FunctionCall, err: FunctionCallException, attempt: int
    ) -> bool:
        """Called when a function call raises an exception.

        By default, this adds a message to the chat telling the LM about the error and allows a retry if the error
        is recoverable and there are remaining retry attempts.

        You may implement an override to customize the error prompt, log the error, or use custom retry logic.
        See :ref:`handle_function_call_exception`.

        :param call: The :class:`.FunctionCall` the model was attempting to make.
        :param err: The error the call raised. Usually this is :class:`.NoSuchFunction` or
            :class:`.WrappedCallException`, although it may be any exception raised by :meth:`do_function_call`.
        :param attempt: The attempt number for the current call (0-indexed).
        :returns: True if the model should retry the call; False if not.
        """
        # log the exception here
        log.debug(f"Call to {call.name} raised an exception: {err}")
        # tell the model what went wrong
        if isinstance(err, NoSuchFunction):
            self.chat_history.append(
                ChatMessage.system(f"The function {err.name!r} is not defined. Only use the provided functions.")
            )
        else:
            # but if it's a user function error, we want to raise it
            log.error(f"Call to {call.name} raised an exception: {err}", exc_info=err)
            self.chat_history.append(ChatMessage.function(call.name, str(err)))

        return attempt < self.retry_attempts and err.retry

    # ==== utility methods ====
    def save(self, fp: PathLike, **kwargs):
        """Save the chat state of this kani to a JSON file. This will overwrite the file if it exists!

        :param fp: The path to the file to save.
        :param kwargs: Additional arguments to pass to Pydantic's ``model_dump_json``.
        """
        data = SavedKani(always_included_messages=self.always_included_messages, chat_history=self.chat_history)
        with open(fp, "w") as f:
            f.write(data.model_dump_json(**kwargs))

    def load(self, fp: PathLike, **kwargs):
        """Load chat state from a JSON file into this kani. This will overwrite any existing chat state!

        :param fp: The path to the file containing the chat state.
        :param kwargs: Additional arguments to pass to Pydantic's ``model_validate_json``.
        """
        with open(fp) as f:
            data = f.read()
        state = SavedKani.model_validate_json(data, **kwargs)
        self.always_included_messages = state.always_included_messages
        self.chat_history = state.chat_history

    # ==== internals ====
    def _auto_truncate_message(self, msg: ChatMessage, max_len: int) -> ChatMessage:
        """Mutate the provided message until it is less than *max_len* tokens long."""
        full_content = msg.content
        if not full_content:
            return msg  # idk how this could happen
        for chunk_divider in ("\n\n", "\n", ". ", ", ", " "):
            # chunk the text
            content = ""
            last_msg = None
            chunks = full_content.split(chunk_divider)
            for idx, chunk in enumerate(chunks):
                # fit in as many chunks as possible
                if idx:
                    content += chunk_divider
                content += chunk
                # when it's too long...
                msg = ChatMessage(role=msg.role, name=msg.name, content=content + "...")
                if self.message_token_len(msg) > max_len:
                    # if we have some text, return it
                    if last_msg:
                        return last_msg
                    # otherwise, we need to split into smaller chunks
                    break
                # otherwise, continue
                last_msg = msg
        # if we get here and have no content, chop it to the first max_len characters
        log.warning(
            "Auto truncate could not find an appropriate place to chunk the text. The returned value will be the first"
            f" {max_len} characters."
        )
        return ChatMessage(role=msg.role, name=msg.name, content=full_content[: max_len - 3] + "...")



from . import engines, exceptions, utils
from .ai_function import AIParam, ai_function
from .kani import Kani
from .models import ChatMessage, ChatRole, FunctionCall
from .utils.cli import chat_in_terminal, chat_in_terminal_async



"""Example from Advanced Usage docs.

This example shows how kani can be served over an API. You'll need to install FastAPI, websockets, and an ASGI server
to run this example::

    pip install fastapi websockets uvicorn

In this example, we implement a simple websocket endpoint that allows a user to chat using WebSockets.
"""
import os

import uvicorn
from fastapi import FastAPI, WebSocket, WebSocketDisconnect

from kani import Kani
from kani.engines.openai import OpenAIEngine

# initialize kani
api_key = os.getenv("OPENAI_API_KEY")
# normally, reusing an OpenAIEngine isn't recommended
# but since we don't define any @ai_functions, it's okay
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")

# initialize FastAPI
app = FastAPI()


@app.websocket("/chat")
async def kani_chat(websocket: WebSocket):
    # accept the websocket and initialize a kani for the connection
    await websocket.accept()
    ai = Kani(engine)
    # take string messages and send string responses
    while True:
        try:
            data = await websocket.receive_text()
            resp = await ai.chat_round_str(data)
            await websocket.send_text(resp)
        # until the client disconnects
        except WebSocketDisconnect:
            return


@app.on_event("shutdown")
async def cleanup_kani():
    """When the application shuts down, cleanly close the kani engine."""
    await engine.close()


if __name__ == "__main__":
    uvicorn.run(app)



"""Run some tests using a real LLaMAv2 model."""
import logging

import pytest

from kani import ChatMessage, Kani
from kani.engines.ctransformers.llama2 import LlamaCTransformersEngine

pytestmark = pytest.mark.llama


@pytest.fixture(scope="module")
def llama():
    return LlamaCTransformersEngine("TheBloke/Llama-2-7B-Chat-GGML", model_file="llama-2-7b-chat.ggmlv3.q4_K_S.bin")


@pytest.fixture()
def create_kani(llama):
    default_system_prompt = (
        "You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while"
        " being safe.  Your answers should not include any harmful, unethical, racist, sexist, toxic,"
        " dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive"
        " in nature.\n\nIf a question does not make any sense, or is not factually coherent, explain why"
        " instead of answering something not correct. If you don't know the answer to a question, please don't"
        " share false information."
    )

    def _inner(system_prompt=default_system_prompt, **kwargs):
        return Kani(llama, system_prompt=system_prompt, **kwargs)

    return _inner


@pytest.fixture(autouse=True, scope="module")
def debug_logging():
    log = logging.getLogger()
    old_level = log.level
    log.setLevel(logging.DEBUG)
    yield
    log.setLevel(old_level)


async def test_llama(create_kani, gh_log):
    """Do one round of conversation with LLaMA."""
    ai = create_kani()
    resp = await ai.chat_round_str("What are some cool things to do in Tokyo?")
    gh_log.write(
        "## LLaMA Basic\n"
        "*These are real outputs from kani running LLaMA v2 on GitHub Actions.*\n"
        "```\n"
        ">>> What are some cool things to do in Tokyo?\n"
    )
    gh_log.write(resp)
    gh_log.write("\n```\n\n")


async def test_chatting_llamas(create_kani, gh_log):
    """Two kanis chatting with each other for 5 rounds."""
    tourist = create_kani(
        "You are a tourist with plans to visit Tokyo.",
        chat_history=[ChatMessage.assistant("What are some cool things to do in Tokyo?")],
    )
    guide = create_kani()

    tourist_response = tourist.chat_history[-1].content
    gh_log.write(
        "## LLaMAs Visit Tokyo\n"
        "*These are real outputs from two kani running LLaMA v2 on GitHub Actions.*\n"
        "```\n"
        f"TOURIST: {tourist_response}\n"
    )
    for _ in range(5):
        guide_response = await guide.chat_round_str(tourist_response)
        tourist_response = await tourist.chat_round_str(guide_response)
        gh_log.write(f"GUIDE: {guide_response}\n\nTOURIST: {tourist_response}\n")
    gh_log.write("```\n\n")



import functools
import os

from kani.ai_function import AIFunction
from kani.exceptions import MissingModelDependencies
from kani.models import ChatMessage
from ..base import BaseEngine
from . import function_calling
from .client import OpenAIClient
from .models import ChatCompletion, FunctionSpec

try:
    import tiktoken
except ImportError as e:
    raise MissingModelDependencies(
        'The OpenAIEngine requires extra dependencies. Please install kani with "pip install kani[openai]".'
    ) from None

# https://platform.openai.com/docs/models
CONTEXT_SIZES_BY_PREFIX = [
    ("gpt-3.5-turbo-16k", 16384),
    ("gpt-3.5-turbo", 4096),
    ("gpt-4-32k", 32768),
    ("gpt-4", 8192),
    # fine-tunes
    ("ft:gpt-3.5-turbo-16k", 16384),
    ("ft:gpt-3.5-turbo", 4096),
    ("ft:gpt-4-32k", 32768),
    ("ft:gpt-4", 8192),
    # completion models
    ("text-davinci-", 4096),
    ("code-", 8000),
    # catch-all
    ("", 2048),  # e.g. aba/babbage/curie/davinci
]


class OpenAIEngine(BaseEngine):
    """Engine for using the OpenAI API.

    This engine supports all chat-based models and fine-tunes.
    """

    def __init__(
        self,
        api_key: str = None,
        model="gpt-3.5-turbo",
        max_context_size: int = None,
        *,
        organization: str = None,
        retry: int = 5,
        api_base: str = "https://api.openai.com/v1",
        headers: dict = None,
        client: OpenAIClient = None,
        **hyperparams,
    ):
        """
        :param api_key: Your OpenAI API key. By default, the API key will be read from the `OPENAI_API_KEY` environment
            variable.
        :param model: The id of the model to use (e.g. "gpt-3.5-turbo", "ft:gpt-3.5-turbo:my-org:custom_suffix:id").
        :param max_context_size: The maximum amount of tokens allowed in the chat prompt. If None, uses the given
            model's full context size.
        :param organization: The OpenAI organization to use in requests (defaults to the API key's default org).
        :param retry: How many times the engine should retry failed HTTP calls with exponential backoff (default 5).
        :param api_base: The base URL of the OpenAI API to use.
        :param headers: A dict of HTTP headers to include with each request.
        :param client: An instance of :class:`.OpenAIClient` (for reusing the same client in multiple engines). You must
            specify exactly one of (api_key, client). If this is passed the ``organization``, ``retry``, ``api_base``,
            and ``headers`` params will be ignored.
        :param hyperparams: Any additional parameters to pass to
            :meth:`.OpenAIClient.create_chat_completion`.
        """
        if api_key and client:
            raise ValueError("You must supply no more than one of (api_key, client).")
        if api_key is None and client is None:
            api_key = os.getenv("OPENAI_API_KEY")
            if api_key is None:
                raise ValueError(
                    "You must supply an `api_key`, `client`, or set the `OPENAI_API_KEY` environment variable to use"
                    " the OpenAIEngine."
                )
        if max_context_size is None:
            max_context_size = next(size for prefix, size in CONTEXT_SIZES_BY_PREFIX if model.startswith(prefix))
        self.client = client or OpenAIClient(
            api_key, organization=organization, retry=retry, api_base=api_base, headers=headers
        )
        self.model = model
        self.max_context_size = max_context_size
        self.hyperparams = hyperparams
        self.tokenizer = None  # tiktoken caches a tokenizer globally in module, so we can unconditionally load it
        self._load_tokenizer()

    def _load_tokenizer(self):
        try:
            self.tokenizer = tiktoken.encoding_for_model(self.model)
        except KeyError:
            self.tokenizer = tiktoken.get_encoding("cl100k_base")

    def message_len(self, message: ChatMessage) -> int:
        mlen = 5  # ChatML = 4, role = 1
        if message.content:
            mlen += len(self.tokenizer.encode(message.content))
        if message.name:
            mlen += len(self.tokenizer.encode(message.name))
        if message.function_call:
            mlen += len(self.tokenizer.encode(message.function_call.name))
            mlen += len(self.tokenizer.encode(message.function_call.arguments))
        return mlen

    async def predict(
        self, messages: list[ChatMessage], functions: list[AIFunction] | None = None, **hyperparams
    ) -> ChatCompletion:
        if functions:
            function_spec = [FunctionSpec(name=f.name, description=f.desc, parameters=f.json_schema) for f in functions]
        else:
            function_spec = None
        completion = await self.client.create_chat_completion(
            model=self.model, messages=messages, functions=function_spec, **self.hyperparams, **hyperparams
        )
        return completion

    def function_token_reserve(self, functions: list[AIFunction]) -> int:
        if not functions:
            return 0
        # wrap an inner impl to use lru_cache with frozensets
        return self._function_token_reserve_impl(frozenset(functions))

    @functools.lru_cache(maxsize=256)
    def _function_token_reserve_impl(self, functions):
        # openai doesn't tell us exactly how their function prompt works, so
        # we rely on community reverse-engineering to build the right prompt
        # hopefully OpenAI releases a utility to calculate this in the future, this seems kind of fragile
        prompt = function_calling.prompt(functions)
        return len(self.tokenizer.encode(prompt)) + 16  # internal MD headers, namespace {} delimiters

    async def close(self):
        await self.client.close()



from typing import Literal

from pydantic import BaseModel

from kani.models import ChatMessage
from ..base import BaseCompletion


# ==== text completions ====
class CompletionLogProbs(BaseModel):
    tokens: list[str]
    token_logprobs: list[float]
    top_logprobs: list[dict[str, float]]
    text_offset: list[int]


class CompletionChoice(BaseModel):
    text: str
    index: int
    logprobs: CompletionLogProbs | None = None
    finish_reason: str | None = None


class CompletionUsage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class Completion(BaseModel):
    id: str
    object: Literal["text_completion"]
    created: int
    model: str
    choices: list[CompletionChoice]
    usage: CompletionUsage

    @property
    def text(self):
        """The text of the top completion."""
        return self.choices[0].text


# ==== chat completions ====
class FunctionSpec(BaseModel):
    name: str
    description: str | None = None
    parameters: dict


class SpecificFunctionCall(BaseModel):
    name: str


# ---- response ----
class ChatCompletionChoice(BaseModel):
    message: ChatMessage
    index: int
    finish_reason: str | None = None


class ChatCompletion(BaseCompletion, BaseModel):
    id: str
    object: Literal["chat.completion"]
    created: int
    model: str
    usage: CompletionUsage
    choices: list[ChatCompletionChoice]

    @property
    def message(self):
        return self.choices[0].message

    @property
    def prompt_tokens(self):
        return self.usage.prompt_tokens

    @property
    def completion_tokens(self):
        # for some reason, the OpenAI API doesn't return the tokens used by ChatML
        # so we add on the length of "<|im_start|>assistant" and "<|im_end|>" here
        return self.usage.completion_tokens + 5

    @property
    def text(self):
        """The text of the most recent chat completion."""
        return self.message.content



import enum
from typing import Annotated, List

from kani import AIParam, ai_function
from kani.ai_function import AIFunction
from .utils import dict_at_least


# ==== setup ====
class EnumS(enum.Enum):
    APPLE = "apple"
    BANANA = "banana"
    COCONUT = "coconut"


class EnumI(enum.Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class EnumI2(enum.IntEnum):
    ONE = 1
    TWO = 2
    THREE = 3


class BadEnum(enum.Enum):
    FOO = 1
    BAR = "two"


# noinspection PyUnusedLocal
async def example_primitives(
    a: str,
    b: float,
    c: bool,
    d: int,
    e: None,
    aa: Annotated[str, AIParam(desc="I am AA")],
    dd: Annotated[int, "I am not an AIParam"] = 2,
):
    """description!"""
    pass


# noinspection PyUnusedLocal
async def example_collections(
    a: list[str],
    b: dict[str, int],
    c: Annotated[list[str], AIParam(desc="I am C")],
    d: Annotated[list[int], "I am not an AIParam"],
    e: List[str],
):
    """collections!"""
    pass


# noinspection PyUnusedLocal
async def example_enums(
    a: EnumS,
    b: EnumI,
    c: EnumI2,
    d: Annotated[EnumS, AIParam(desc="I am D")],
    e: Annotated[EnumI, "I am not an AIParam"],
):
    """enums!"""
    pass


def create_ai_function(f):
    decorated = ai_function(f)
    return AIFunction(decorated, **decorated.__ai_function__)


# ==== tests ====
def test_schema_primitives():
    f = create_ai_function(example_primitives)
    assert isinstance(f, AIFunction)
    assert dict_at_least(
        f.json_schema,
        {
            "properties": {
                "a": {"type": "string"},
                "b": {"type": "number"},
                "c": {"type": "boolean"},
                "d": {"type": "integer"},
                "e": {"type": "null"},
                "aa": {"description": "I am AA", "type": "string"},
                "dd": {"default": 2, "type": "integer"},
            },
            "required": ["a", "b", "c", "d", "e", "aa"],
            "type": "object",
        },
    )


def test_schema_collections():
    f = create_ai_function(example_collections)
    assert isinstance(f, AIFunction)
    assert dict_at_least(
        f.json_schema,
        {
            "properties": {
                "a": {"items": {"type": "string"}, "type": "array"},
                "b": {"additionalProperties": {"type": "integer"}, "type": "object"},
                "c": {"description": "I am C", "items": {"type": "string"}, "type": "array"},
                "d": {"items": {"type": "integer"}, "type": "array"},
                "e": {"items": {"type": "string"}, "type": "array"},
            },
            "required": ["a", "b", "c", "d", "e"],
            "type": "object",
        },
    )


def test_schema_enums():
    f = create_ai_function(example_enums)
    assert isinstance(f, AIFunction)
    assert dict_at_least(
        f.json_schema,
        {
            "properties": {
                "a": {"enum": ["apple", "banana", "coconut"], "type": "string"},
                "b": {"enum": [1, 2, 3], "type": "integer"},
                "c": {"enum": [1, 2, 3], "type": "integer"},
                "d": {"enum": ["apple", "banana", "coconut"], "type": "string", "description": "I am D"},
                "e": {"enum": [1, 2, 3], "type": "integer"},
            },
            "required": ["a", "b", "c", "d", "e"],
            "type": "object",
        },
    )



from kani.ai_function import AIFunction
from kani.exceptions import MissingModelDependencies
from kani.models import ChatMessage, ChatRole
from .base import HuggingEngine

try:
    import sentencepiece
except ImportError:
    raise MissingModelDependencies(
        'The VicunaEngine requires extra dependencies. Please install kani with "pip install'
        " 'kani[huggingface,llama]'\". You will also need to install PyTorch manually."
    ) from None

try:
    import accelerate

    _low_cpu_mem_usage = True
except ImportError:
    _low_cpu_mem_usage = False


class VicunaEngine(HuggingEngine):
    """Implementation of Vicuna (a LLaMA v1 fine-tune) using huggingface transformers.

    You may also use the 13b, 33b, or other LLaMA models that use the Vicuna prompt by passing the HuggingFace model
    ID to the initializer.

    .. seealso:: https://huggingface.co/lmsys/vicuna-7b-v1.3

    .. tip:: See :ref:`4b_quant` for information about loading a quantized model for lower memory usage.

    .. code-block:: python

        engine = VicunaEngine("lmsys/vicuna-7b-v1.3")
        ai = Kani(engine)
    """

    # all prompts start with a hidden <s> token and ASSISTANT:
    token_reserve = 7
    """The Vicuna prompt starts with a hidden <s> token and "ASSISTANT:", which uses 7 tokens."""

    def __init__(self, model_id: str = "lmsys/vicuna-7b-v1.3", *args, **kwargs):
        tokenizer_kwargs = kwargs.pop("tokenizer_kwargs", {})
        tokenizer_kwargs.setdefault("use_fast", False)

        model_load_kwargs = kwargs.pop("model_load_kwargs", {})
        model_load_kwargs.setdefault("low_cpu_mem_usage", _low_cpu_mem_usage)

        kwargs.setdefault("max_context_size", 2048)  # LLaMA has 2048 token window
        super().__init__(
            model_id, *args, tokenizer_kwargs=tokenizer_kwargs, model_load_kwargs=model_load_kwargs, **kwargs
        )

    def build_prompt(self, messages: list[ChatMessage], functions: list[AIFunction] | None = None) -> str:
        prompt_lines = []
        for message in messages:
            if message.role == ChatRole.USER:
                prompt_lines.append(f"USER: {message.content}")
            elif message.role == ChatRole.ASSISTANT:
                prompt_lines.append(f"ASSISTANT: {message.content}</s>")
            else:
                prompt_lines.append(f"{message.content}\n")
        prompt = "\n".join(prompt_lines)
        return f"{prompt}\nASSISTANT:"

    def message_len(self, message: ChatMessage) -> int:
        # https://github.com/lm-sys/FastChat/blob/main/docs/vicuna_weights_version.md#example-prompt-weights-v11-and-v13
        # remove 1 for the <s> token at the start
        if message.role == ChatRole.USER:
            # USER: {}\n -> 5
            return self.tokenizer(message.content, return_length=True).length + 4
        elif message.role == ChatRole.ASSISTANT:
            # ASSISTANT: {}</s>\n -> 8
            return self.tokenizer(message.content, return_length=True).length + 7
        # {}\n\n -> 2
        return self.tokenizer(message.content, return_length=True).length + 1



import inspect
import typing
from typing import TYPE_CHECKING, Optional

import pydantic

if TYPE_CHECKING:
    from .ai_function import AIParam

# this is the same as Pydantic's as of v2.1, but we specify it here because some downstream things rely on it
# (e.g. engines.openai.function_calling_tokens)
REF_TEMPLATE = "#/$defs/{model}"


class AIParamSchema:
    """Used to annotate parameters of AIFunctions in order to make generating their schema nicer.

    This class is only used internally within kani and generally shouldn't be constructed manually.
    """

    def __init__(self, name: str, t: type, default, aiparam: Optional["AIParam"] = None):
        self.name = name
        self.type = t
        self.default = default
        self.aiparam = aiparam

    @property
    def required(self):
        return self.default is inspect.Parameter.empty

    @property
    def origin_type(self):
        """If the type takes parameters (e.g. list[...]), the base type (i.e. list). Otherwise same as the type."""
        return typing.get_origin(self.type) or self.type

    @property
    def description(self):
        return self.aiparam.desc if self.aiparam is not None else None


class JSONSchemaBuilder(pydantic.json_schema.GenerateJsonSchema):
    """Subclass of the Pydantic JSON schema builder to provide more fine-grained control over titles and refs."""

    def field_title_should_be_set(self, _) -> bool:
        # We never want titles to be set.
        return False

    def _update_class_schema(self, json_schema, title, *args, **kwargs):
        # don't set title, transform it to None
        return super()._update_class_schema(json_schema, None, *args, **kwargs)

    def _build_definitions_remapping(self):
        # we need to remember what remappings Pydantic did so we can flatten them later
        self.remapping = super()._build_definitions_remapping()
        return self.remapping

    def flatten_singleton_refs(self, json_schema, threshold=1):
        """Substitute any refs that only occur once with the literal."""
        defs = json_schema.get("$defs")
        if defs is None:
            return json_schema
        def_counts = self.get_json_ref_counts(json_schema)

        def _flatten(obj):
            if not isinstance(obj, dict):
                return obj

            new_obj = {}
            for k, v in obj.items():
                if isinstance(v, list):
                    new_obj[k] = [_flatten(x) for x in v]
                elif isinstance(v, dict):
                    new_obj[k] = _flatten(v)
                elif k == "$ref" and def_counts[v] <= threshold:
                    new_obj.update(self.get_schema_from_definitions(v))
                else:
                    new_obj[k] = v
            return new_obj

        json_schema = _flatten(json_schema)

        self._garbage_collect_definitions(json_schema)
        if self.definitions:
            json_schema["$defs"] = self.definitions
        else:
            json_schema.pop("$defs")
        return json_schema

    def flatten_singleton_allof(self, json_schema, flatten_refs=False):
        """Bring any allOf that only have a single child up to the sibling level."""

        def _flatten(obj):
            if not isinstance(obj, dict):
                return obj

            for k, v in obj.items():
                if isinstance(v, list):
                    obj[k] = [_flatten(x) for x in v]
                elif isinstance(v, dict):
                    if "allOf" in v and len(v["allOf"]) == 1:
                        entry = v["allOf"][0]
                        if "$ref" not in entry or flatten_refs:
                            v.pop("allOf")
                            v.update(entry)
                    obj[k] = _flatten(v)
            return obj

        return _flatten(json_schema)

    def remove_titles(self, json_schema):
        """Recursively remove title tags from the schema in place."""
        if not isinstance(json_schema, dict):
            return
        # only kill the title key if it's a str; a prop could be named "title"
        if "title" in json_schema and isinstance(json_schema["title"], str):
            json_schema.pop("title")
        # then recurse
        for k, v in json_schema.items():
            if isinstance(v, list):
                for x in v:
                    self.remove_titles(x)
            elif isinstance(v, dict):
                self.remove_titles(v)

    def generate(self, *args, **kwargs):
        json_schema = super().generate(*args, **kwargs)
        # take the remappings and make it canonical
        new_json_to_defs_refs = {
            new: self.remapping.defs_remapping[self.json_to_defs_refs[old]]
            for old, new in self.remapping.json_remapping.items()
        }
        self.json_to_defs_refs = new_json_to_defs_refs
        new_definitions = {new: self.definitions[old] for old, new in self.remapping.defs_remapping.items()}
        self.definitions = new_definitions
        # flatten any singleton def/refs
        json_schema = self.flatten_singleton_refs(json_schema, 2)
        # flatten singleton allOf
        json_schema = self.flatten_singleton_allof(json_schema)
        # recursively remove title tags
        self.remove_titles(json_schema)
        return json_schema


def create_json_schema(params: list[AIParamSchema]) -> dict:
    """Create a JSON schema from a list of parameters to an AIFunction.

    There are some subtle differences compared to how Pydantic creates JSON schemas by default; most notably,
    some titles are omitted to minimize the token count, and defaults are not exposed.
    """
    # create pydantic fields for each AIParam
    fields = {}
    for param in params:
        field_kwargs = dict(description=param.description)
        if not param.required:
            field_kwargs["default"] = param.default
        fields[param.name] = (param.type, pydantic.Field(**field_kwargs))
    # create a temp model for generating json schemas
    pydantic_model = pydantic.create_model("_FunctionSpec", **fields)
    return pydantic_model.model_json_schema(schema_generator=JSONSchemaBuilder, ref_template=REF_TEMPLATE)



"""FP4 quantization compresses models' weights to a 4-bit representation, saving memory on the GPU.

This example shows how to initialize a HuggingFace engine with fp4 quantization enabled on the GPU.

You will need to install ``bitsandbytes`` and ``accelerate`` from pip.
"""
import asyncio
import time

from kani import Kani, chat_in_terminal
from kani.engines.huggingface.llama2 import LlamaEngine

engine = LlamaEngine(
    use_auth_token=True,
    strict=True,
    model_load_kwargs={
        "device_map": "auto",
        "load_in_4bit": True,
    },
)
ai = Kani(
    engine,
    system_prompt=(
        "You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe."
        " Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content."
        " Please ensure that your responses are socially unbiased and positive in nature.\n\nIf a question does not"
        " make any sense, or is not factually coherent, explain why instead of answering something not correct. If you"
        " don't know the answer to a question, please don't share false information."
    ),
)


async def time_completion():
    before = time.monotonic()
    message = await ai.chat_round("What are some interesting things to do in Tokyo?", top_k=1, do_sample=True)
    print(message.content)
    print(f"Tokens: {ai.message_token_len(message)}")
    after = time.monotonic()
    print(f"Time: {after - before}")


if __name__ == "__main__":
    chat_in_terminal(ai)
    # asyncio.run(time_completion())



"""Example from the Customization docs.

This example shows how to override kani's function call handler to add instrumentation.
"""
import collections
import datetime
import os

from kani import Kani, ai_function, chat_in_terminal
from kani.engines.openai import OpenAIEngine
from kani.exceptions import FunctionCallException

api_key = os.getenv("OPENAI_API_KEY")
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")


class TrackCallsKani(Kani):
    # You can overload __init__ and track kani-specific state:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.successful_calls = collections.Counter()
        self.failed_calls = collections.Counter()

    async def do_function_call(self, call):
        try:
            result = await super().do_function_call(call)
            self.successful_calls[call.name] += 1
            return result
        except FunctionCallException:
            self.failed_calls[call.name] += 1
            raise

    # Let's give the model some functions to work with:
    @ai_function()
    def get_time(self):
        """Get the current time in the user's time zone."""
        # oh no! the clock is broken!
        raise RuntimeError("The time API is currently offline. Please try using `get_date_and_time`.")

    @ai_function()
    def get_date_and_time(self):
        """Get the current day and time in the user's time zone."""
        return str(datetime.datetime.now())


ai = TrackCallsKani(engine)
if __name__ == "__main__":
    chat_in_terminal(ai, rounds=1)
    print("Successful:", ai.successful_calls)
    print("Failed:", ai.failed_calls)



import os

from kani import Kani, chat_in_terminal
from kani.engines.openai import OpenAIEngine

# Replace this with your OpenAI API key (https://platform.openai.com/account/api-keys)
# or run `export OPENAI_API_KEY="sk-..."` in your terminal.
api_key = os.getenv("OPENAI_API_KEY")

# kani uses an Engine to interact with the language model. You can specify other model parameters here,
# like temperature=0.7.
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")

# The kani manages the chat state, prompting, and function calling. Here, we only give it the engine to call
# ChatGPT, but you can specify other parameters like system_prompt="You are..." here.
ai = Kani(engine)

# kani comes with a utility to interact with a kani through your terminal! Check out the docs for how to use kani
# programmatically.
if __name__ == "__main__":
    chat_in_terminal(ai)



"""Example from the Customization docs.

This example shows how to override kani's function call exception handler to use a custom prompt.
"""
import os

from kani import ChatMessage, Kani, ai_function, chat_in_terminal
from kani.engines.openai import OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")


class CustomExceptionPromptKani(Kani):
    async def handle_function_call_exception(self, call, err, attempt):
        self.chat_history.append(
            ChatMessage.system(
                f"The call encountered an error. Relay this error message to the user in a sarcastic manner: {err}"
            )
        )
        return attempt < self.retry_attempts and err.retry

    @ai_function()
    def get_time(self):
        """Get the current time in the user's time zone."""
        raise RuntimeError("The time API is currently offline (error 0xDEADBEEF).")


ai = CustomExceptionPromptKani(engine)
if __name__ == "__main__":
    chat_in_terminal(ai)



"""Model-agnostic classes used to represent the chat state and function calls."""
import enum
import json

from pydantic import BaseModel, ConfigDict


# ==== chat ====
class ChatRole(enum.Enum):
    """Represents who said a chat message."""

    SYSTEM = "system"
    """The message is from the system (usually a steering prompt)."""

    USER = "user"
    """The message is from the user."""

    ASSISTANT = "assistant"
    """The message is from the language model."""

    FUNCTION = "function"
    """The message is the result of a function call."""


class FunctionCall(BaseModel):
    """Represents a model's request to call a function."""

    model_config = ConfigDict(frozen=True)

    name: str
    """The name of the requested function."""

    arguments: str
    """The arguments to call it with, encoded in JSON."""

    @property
    def kwargs(self) -> dict:
        """The arguments to call the function with, with JSON decoded to a Python dict."""
        return json.loads(self.arguments)

    @classmethod
    def with_args(cls, name: str, **kwargs):
        """Create a function call with the given arguments (e.g. for few-shot prompting)."""
        return cls(name=name, arguments=json.dumps(kwargs))


class ChatMessage(BaseModel):
    """Represents a message in the chat context."""

    model_config = ConfigDict(frozen=True)

    role: ChatRole
    """Who said the message?"""

    content: str | None
    """The content of the message. Can be None only if the message is a requested function call from the assistant."""

    name: str | None = None
    """The name of the user who sent the message, if set (user messages only)."""

    function_call: FunctionCall | None = None
    """The function requested by the model (assistant messages only)."""

    @classmethod
    def system(cls, content: str, **kwargs):
        """Create a new system message."""
        return cls(role=ChatRole.SYSTEM, content=content, **kwargs)

    @classmethod
    def user(cls, content: str, **kwargs):
        """Create a new user message."""
        return cls(role=ChatRole.USER, content=content, **kwargs)

    @classmethod
    def assistant(cls, content: str | None, **kwargs):
        """Create a new assistant message."""
        return cls(role=ChatRole.ASSISTANT, content=content, **kwargs)

    @classmethod
    def function(cls, name: str, content: str, **kwargs):
        """Create a new function message."""
        return cls(role=ChatRole.FUNCTION, content=content, name=name, **kwargs)



import abc
import logging

import aiohttp

from ..exceptions import HTTPException, HTTPStatusException, HTTPTimeout


class BaseClient(abc.ABC):
    """aiohttp-based HTTP client to help implement HTTP-based engines."""

    SERVICE_BASE: str
    """The base route of the HTTP API."""

    logger: logging.Logger

    def __init__(self, http: aiohttp.ClientSession = None):
        """
        :param http: The :class:`aiohttp.ClientSession` to use; if not provided, creates a new session.
        """
        self.http = http

    def __init_subclass__(cls, **kwargs):
        # to initialize the logger with the right name, set it when the subclass is initialized
        # this prevents all loggers logging as kani.engines.httpclient
        cls.logger = logging.getLogger(cls.__module__)

    async def request(self, method: str, route: str, response_as_text=False, **kwargs):
        """Makes an HTTP request to the given route (relative to the base route).

        :param method: The HTTP method to use (e.g. 'GET', 'POST').
        :param route: The route to make the request to (relative to the ``SERVICE_BASE``).
        :param response_as_text: If True, return the response's content as a string; otherwise decodes it as JSON.
        :raises HTTPStatusException: The request returned a non-2xx response.
        :raises HTTPTimeout: The request timed out.
        :raises HTTPException: The response could not be deserialized.
        """
        if self.http is None:
            self.http = aiohttp.ClientSession()
        try:
            async with self.http.request(method, f"{self.SERVICE_BASE}{route}", **kwargs) as resp:
                self.logger.debug(f"{method} {self.SERVICE_BASE}{route} returned {resp.status}")
                if not 199 < resp.status < 300:
                    data = await resp.text()
                    self.logger.warning(
                        f"{method} {self.SERVICE_BASE}{route} returned {resp.status} {resp.reason}\n{data}"
                    )
                    raise HTTPStatusException(resp.status, f"Request returned an error: {resp.status}: {resp.reason}")
                try:
                    if not response_as_text:
                        data = await resp.json()
                    else:
                        data = await resp.text()
                    self.logger.debug(data)
                except (aiohttp.ContentTypeError, ValueError, TypeError) as e:
                    data = await resp.text()
                    self.logger.warning(
                        f"{method} {self.SERVICE_BASE}{route} response could not be deserialized:\n{data}"
                    )
                    raise HTTPException(f"Could not deserialize response: {data}") from e
        except aiohttp.ServerTimeoutError as e:
            self.logger.warning(f"Request timeout: {method} {self.SERVICE_BASE}{route}")
            raise HTTPTimeout() from e
        return data

    async def get(self, route: str, **kwargs):
        """Convenience method; equivalent to ``self.request("GET", route, **kwargs)``."""
        return await self.request("GET", route, **kwargs)

    async def post(self, route: str, **kwargs):
        """Convenience method; equivalent to ``self.request("POST", route, **kwargs)``."""
        return await self.request("POST", route, **kwargs)

    async def close(self):
        """Close the underlying aiohttp session."""
        if self.http is None:
            return
        await self.http.close()
        self.http = None  # this allows us to reuse the client after it has been closed



import os

import d20  # pip install d20

from kani import Kani, ai_function, chat_in_terminal
from kani.engines.openai import OpenAIEngine

engine = OpenAIEngine(os.getenv("OPENAI_API_KEY"), model="gpt-3.5-turbo")


class DiceKani(Kani):
    @ai_function
    def roll(self, dice: str):
        """Roll some dice or do math. Dice should be specified in the XdY format."""
        return d20.roll(dice).result


ai = DiceKani(engine)

if __name__ == "__main__":
    chat_in_terminal(ai)



"""The CLI utilities allow you to play with a chat session directly from a terminal."""
import asyncio
import logging
import os

from kani.kani import Kani
from kani.models import ChatMessage


def _function_formatter(message: ChatMessage):
    return f"Thinking ({message.function_call.name})..."


async def chat_in_terminal_async(kani: Kani, rounds: int = 0, stopword: str = None):
    """Async version of :func:`.chat_in_terminal`.
    Use in environments when there is already an asyncio loop running (e.g. Google Colab).
    """
    if os.getenv("KANI_DEBUG") is not None:
        logging.basicConfig(level=logging.DEBUG)

    try:
        round_num = 0
        while round_num < rounds or not rounds:
            round_num += 1
            query = input("USER: ")
            if stopword and query == stopword:
                break
            async for msg in kani.full_round_str(query, function_call_formatter=_function_formatter):
                print(f"AI: {msg}")
    except KeyboardInterrupt:
        pass
    finally:
        await kani.engine.close()


def chat_in_terminal(kani: Kani, rounds: int = 0, stopword: str = None):
    """Chat with a kani right in your terminal.

    Useful for playing with kani, quick prompt engineering, or demoing the library.

    If the environment variable ``KANI_DEBUG`` is set, debug logging will be enabled.

    .. warning::

        This function is only a development utility and should not be used in production.

    :param rounds: The number of chat rounds to play (defaults to 0 for infinite).
    :param stopword: Break out of the chat loop if the user sends this message.
    """
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        pass
    else:
        print(
            f"WARNING: It looks like you're in an environment with a running asyncio loop (e.g. Google Colab).\nYou"
            f" should use `await chat_in_terminal_async(...)` instead."
        )
        return
    asyncio.run(chat_in_terminal_async(kani, rounds=rounds, stopword=stopword))



"""
This example shows how to use kani's main entrypoints by setting up an async context.

In an experimental or production setting, you would likely load in data or accept user input through your own
mechanisms, rather than ``input()``. This example shows how the synchronous ``chat_in_terminal`` works under the hood
(with some simplifications, of course).

If you haven't already, go take a look at `1_quickstart.py` for an explanation of how to set up kani first.
"""
import asyncio

from kani import Kani
from kani.engines.openai import OpenAIEngine

api_key = "sk-..."
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")
ai = Kani(engine, system_prompt="You are a helpful assistant.")


# define your function normally, using `async def` instead of `def`
async def chat_with_kani():
    while True:
        user_message = input("USER: ")
        # now, you can use `await` to call kani's async methods
        message = await ai.chat_round_str(user_message)
        print("AI:", message)


# use `asyncio.run` to call your async function to start the program
asyncio.run(chat_with_kani())



"""Example from Advanced Usage docs.

This example shows how kani can be used in conjunction with function calling to spawn sub-kani. For example, you might
give a sub-kani a different set of functions, use a more powerful engine, or perform an isolated task.

In this example, we show how a sub-kani can use a different engine to perform summarization, then give the results
to the parent kani.
"""
import os

from kani import Kani, ai_function, chat_in_terminal
from kani.engines.openai import OpenAIClient, OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")

fast_model = "gpt-3.5-turbo"
long_context_model = "gpt-3.5-turbo-16k"


class KaniWithAISummarization(Kani):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # save a single client here so we can reuse it across multiple calls
        self.openai_client = OpenAIClient(api_key)

    @ai_function()
    async def summarize_conversation(self):
        """Get the summary of the conversation so far."""
        # in this AI Function, we can spawn a sub-kani with a model that can handle
        # longer contexts, since the conversation may be longer than the fast model's
        # context window
        long_context_engine = OpenAIEngine(client=self.openai_client, model=long_context_model)
        # first, copy the parent's chat history to the child, except the last user message
        # and the function call ([:-2])
        sub_kani = Kani(long_context_engine, chat_history=self.chat_history[:-2])
        # then we ask it to summarize the whole thing, and return the result to the parent kani
        return await sub_kani.chat_round_str("Please summarize the conversation so far.")


fast_engine = OpenAIEngine(api_key, fast_model)
ai = KaniWithAISummarization(fast_engine)
if __name__ == "__main__":
    chat_in_terminal(ai)



import os

from pydantic import BaseModel

from ..models import ChatMessage

PathLike = str | bytes | os.PathLike


class SavedKani(BaseModel):
    version: int = 1
    always_included_messages: list[ChatMessage]
    chat_history: list[ChatMessage]



"""
This example shows how you can set the chat_history to accomplish few-shot prompting with kani.

If you haven't already, go take a look at `1_quickstart.py` for an explanation of how to set up kani.
"""
import os

from kani import ChatMessage, Kani, chat_in_terminal
from kani.engines.openai import OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")

# create a few-shot prompt by providing query-answer pairs
fewshot = [
    ChatMessage.user("thank you"),
    ChatMessage.assistant("arigato"),
    ChatMessage.user("good morning"),
    ChatMessage.assistant("ohayo"),
]
# then supply it to kani when you initialize
ai = Kani(engine, chat_history=fewshot)

if __name__ == "__main__":
    chat_in_terminal(ai)
    # USER: crab
    # ASSISTANT: kani



"""Example from the Customization docs.

This example shows how to override kani's chat prompt building functionality.
"""
import os

from kani import Kani, chat_in_terminal
from kani.engines.openai import OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")


class LastFourKani(Kani):
    async def get_prompt(self):
        """
        Only include the most recent 4 messages (omitting earlier ones to fit in the token length if necessary)
        and any always included messages.
        """
        # calculate how many tokens we have for the prompt, accounting for the system prompt, always_included_messages,
        # any tokens reserved by the engine, and the response
        remaining = self.max_context_size - self.always_len
        # working backwards through history...
        messages = []
        for message in reversed(self.chat_history[-4:]):
            # if the message fits in the space we have remaining...
            message_len = self.message_token_len(message)
            remaining -= message_len
            if remaining > 0:
                # add it to the returned prompt!
                messages.insert(0, message)
            else:
                break
        return self.always_included_messages + messages


ai = LastFourKani(engine)
if __name__ == "__main__":
    chat_in_terminal(ai)
